import json
from pathlib import Path


#              CMAC v1.1L does not correct for the interactive effects of Atm-I and 
#              SZA, which can be significant. Consideration of such effects and others
#              will be standard in future commercial versions of CMAC.
#
#              CMAC v1.1L is not approved for use with data other than that acquired by
#              Landsat 8/9. Calibration of CMAC for other satellite imaging systems is in
#              progress. 


def read_config(cfg_filename):

    # Initializations go here.
    cfg_param_dict = {}
    err_flag = False
    err_msg = ''

    # Get the actual user-supplied info. The rest of
    # the configuration info, at least for CMAC V1.1
    # is "hard-coded"; future versions of this software
    # could add these to the configuration file (as in
    # earlier CMAC development).
    if not Path(cfg_filename).exists():
        err_msg = f'ERROR: configuration file "{cfg_filename}" not found'
        print(err_msg)
        return {}

    # Read the config file.
    with open(cfg_filename, 'r') as fid:
        json_cfg_data = json.load(fid)

    # Get the path information.
    root_toa_path = json_cfg_data["toaRootDrive"]
    root_cmac_path = json_cfg_data["cmacRootDrive"]

    # Get the particular satellite/sensor configuration
    satellite = json_cfg_data["satellite"]
    sensor = json_cfg_data["sensor"]

    # Get information to locate imagery on a local filesystem.
    #
    additional_path_info = json_cfg_data["additionalPathInfo"]
    
    # Get the acquisition information represented in the product ID.
    product_id = json_cfg_data["productID"]

    option1_folder = json_cfg_data["option1Folder"]
    option2_folder = json_cfg_data["option2Folder"]

    # Input (and output) file extensions.
    file_ext = json_cfg_data["fileExt"]

    # Get the Atm-I related options.
    nodata_value = json_cfg_data["nodataValue"]
    atmi_gridcell_dim = json_cfg_data["atmiGridcellDim"]
    
    median_filter_on = json_cfg_data["medianFilterOn"]
    median_filter_radius = json_cfg_data["medianFilterRadius"]
    gaussian_filter_on = json_cfg_data["gaussianFilterOn"]
    gaussian_filter_radius = json_cfg_data["gaussianFilterRadius"]

    atmi_cap_on = json_cfg_data["atmiCapOn"]
    atmi_cap_level = json_cfg_data["atmiCapLevel"]



    # Populate the configuration parameter data structure:

    # Imaging system info.
    cfg_param_dict.update({"satellite" : satellite})
    cfg_param_dict.update({"sensor" : sensor})

    # Path and file info.
    cfg_param_dict.update({"root_toa_path" : root_toa_path})
    cfg_param_dict.update({"root_cmac_path" : root_cmac_path})
    cfg_param_dict.update({"option1_folder" : option1_folder})
    cfg_param_dict.update({"option2_folder" : option2_folder})
    cfg_param_dict.update({"file_ext" : file_ext})

    cfg_param_dict.update({"additional_path_info" : additional_path_info})
    cfg_param_dict.update({"product_id" : product_id})

    # CMAC version info.
    cfg_param_dict.update({"cmac_cal_version" : ""})
    cfg_param_dict.update({"cmac_software_version" : "PyCMAC_V11L"})

    # General processing and Atm-I related info.
    cfg_param_dict.update({"nodata_value" : nodata_value})
    cfg_param_dict.update({"bands_for_atmi_model" : [2, 4]})
    cfg_param_dict.update({"bands_to_correct" : [2, 3, 4, 5]})

    if atmi_cap_on == 1:
        print('Atm-I capping enabled')
        cfg_param_dict.update({"atmi_cap_on" : True})
        print(f'Atm-I cap level:         {atmi_cap_level}')
        cfg_param_dict.update({"atmi_cap_level" : atmi_cap_level})
        
    else:
        print("Atm-I capping disabled")
        cfg_param_dict.update({"atmi_cap_on" : False})
        cfg_param_dict.update({"atmi_cap_level" : 0})
        
    if median_filter_on == 1:
        print('Atm-I median filter smoothing enabled')
        cfg_param_dict.update({"median_filter_on" : True})
        print(f'Median filter radius:    {median_filter_radius}')
        cfg_param_dict.update({"median_filter_radius" : median_filter_radius})
    else:
        print('Median filter smoothing disabled')
        cfg_param_dict.update({"median_filter_on" : False})
        cfg_param_dict.update({"median_filter_radius" : 0})

    if gaussian_filter_on == 1:
        print('Gaussian filter smoothing enabled')
        cfg_param_dict.update({"gaussian_filter_on" : True})
        print(f'Gaussian filter radius:   {gaussian_filter_radius}')
        cfg_param_dict.update({"gaussian_filter_radius" : gaussian_filter_radius})
    else:
        print('Gaussian filter smoothing disabled')
        cfg_param_dict.update({"gaussian_filter_on" : False})
        cfg_param_dict.update({"gaussian_filter_radius" : 0})

    # Atm-I model and gridcell dimensions.
    cfg_param_dict.update({"atmi_model_coeffs" : []})
    cfg_param_dict.update({"atmi_gridcell_size" : []})

    cfg_param_dict.update({"mpc_scaling_gain" : 1.0})
    cfg_param_dict.update({"mpc_scaling_offset" : 0.0})

 
    # Calibration model parameter info. Assumes, expects to process 4 bands;
    # models for other bands can be added.
    cfg_param_dict.update({"cal_offset_models" : [[], [], [], []]})
    cfg_param_dict.update({"cal_slope_models" : [[], [], [], []]})


    return cfg_param_dict
