from pathlib import Path

import numpy as np

import cmac_gdal_io as io
import cmac_metadata as meta



def convert_toa_band(band_image_filename, band_reflectance_conversion_factors,
                     nodata_value, scale_factor, projection_info, 
                     geotransform_info, output_path, image_basename):

    # Make sure the converted image file doesn't already exist. If it does,
    # skip it.
    band_id = band_reflectance_conversion_factors[0]
    converted_image_fname = output_path + '/' + image_basename + f'_{band_id}.tif'
    if Path(converted_image_fname).exists():
        print(f'{band_id} image already converted, skipping...')
        return converted_image_fname


    # Open the image file.
    print(f'Converting band {band_id} image file...')
    image_data_buffer = io.read_image_from_file(band_image_filename)
    if image_data_buffer is None:
        err_msg = f'ERROR: cannot load {band_image_filename}'
        print(err_msg)
        return ''

    # Get the proper reflectance conversion factors.
    refl_mult_factor = band_reflectance_conversion_factors[1]
    refl_add_factor = band_reflectance_conversion_factors[2]

    # Get the 'nodata' fill region.
    pix_nodata = (image_data_buffer == nodata_value)

    # Apply the conversion.
    out_data_buffer = np.round(scale_factor * ((refl_mult_factor * image_data_buffer) \
       + refl_add_factor), 0).astype(np.uint16)

    # Re-establish the nodata fill.
    out_data_buffer[pix_nodata] = nodata_value

    # Save the converted image. The band label is the first parameter in the
    # 'band_reflectance_conversion_factors' input.
    out_image_basename = image_basename + f'_{band_id}'
    status = io.write_image_to_geotiff(out_data_buffer,  projection_info,
                                       geotransform_info, output_path,
                                       out_image_basename)
    if status < 0:
        err_msg = f'ERROR: cannot save converted band {band_id} image'
        print(err_msg)
        return ''

    return converted_image_fname

    


