
import numpy as np

import cmac_gdal_io as io



def correct_band_toa(band_toa_image_buffer, atmi_buffer,
                     band_mask_image_buffer, band_num_img_rows,
                     band_num_img_cols, atmi_cap_on,
                     atmi_cap_level, nodata_value,
                     band_slope_model_coeffs,
                     band_offset_model_coeffs,
                     uint16_max_val):

    # Initializations.
    num_slope_coeffs = len(band_slope_model_coeffs)
    num_offset_coeffs = len(band_offset_model_coeffs)

    # In python, it's generally more "efficient" computationally
    # to "vectorize" calculations.

    # Determine the pixels that are at the specified 'nodata'
    # value.
    toa_nodata_pix = (band_toa_image_buffer == nodata_value)

    # Determine all pixels where the Atm-I value is above the cap
    # level. Make sure to exclude all pixels that aren't at the
    # specified 'nodata' pixel value.
    final_atmi_buffer = atmi_buffer
    pix_above_atmi_cap \
        = ~toa_nodata_pix & (atmi_buffer > atmi_cap_level)
    if (atmi_cap_on):
        final_atmi_buffer[pix_above_atmi_cap] = atmi_cap_level

    # Determine all pixels where the Atm-I value is at or below
    # the specified "hardstop" level. Pixels with such low levels
    # of Atm-I should not (in the current view) be corrected,
    # 1) as the TOA pixel value is very close to the "true"
    # surface reflectance value (i.e. very "clean" atmosphere);
    # and/or 2) there are significant issues with the Atm-I
    # model that are causing this low estimate.
    

    # Generate slope and offset raster image buffers.
    # WARNING: Using numpy's polyval() function/method to evaluate
    # the calibration slope and offset models requires the highest
    # order term in the polynomial definition to be first. This
    # code has coefficients defined in the reverse order,
    # in order to work with an implementation of Horner's polynomial
    # evaluation algorithm used in the C++ implementation of this
    # software.
    cal_offset_model_coeffs = np.flip(np.array(band_offset_model_coeffs))
    cal_slope_model_coeffs = np.flip(np.array(band_slope_model_coeffs))

    cal_offset_buffer = np.polyval(cal_offset_model_coeffs, final_atmi_buffer)
    cal_slope_buffer = np.polyval(cal_slope_model_coeffs, final_atmi_buffer)

    # Apply the CMAC correction.
    sr_image_buffer \
        = ((band_toa_image_buffer - cal_offset_buffer) \
            / (cal_slope_buffer + 1.))

    # Apply the nodata mask.
    sr_image_buffer = sr_image_buffer * band_mask_image_buffer

    # Reset pixels with Atm-I at or below the band "hardstop" level to TOA values.
    
    # Determine the number of pixels that over-corrected to a negative value,
    # corrected to a value between 0 and 1, or corrected above 10000 (representing
    # a scaled surface reflectance value > 1.0)
    sr_pix_negative = ~toa_nodata_pix & (sr_image_buffer < 0)
    sr_pix_below_one = ~toa_nodata_pix & (sr_image_buffer < 1)
    sr_pix_above_10k = ~toa_nodata_pix & (sr_image_buffer > 10000.)

    num_negative_corrects = np.sum(sr_pix_negative)
    num_below_one_corrects = np.sum(sr_pix_below_one)
    num_above_10k_corrects = np.sum(sr_pix_above_10k)

    # Reset pixels that corrected negative or below 1.0 to 1.0 and/or
    # pixels that corrected above 10k to 10k.
    sr_image_buffer[sr_pix_negative] = 1
    sr_image_buffer[sr_pix_below_one] = 1
    sr_image_buffer[sr_pix_above_10k] = 10000

    # Round the corrected image to unsigned 16-bit
    # integer.
    sr_image_buffer = np.round(sr_image_buffer, 0).astype(np.uint16)

    # Should be done.
    return sr_image_buffer, num_negative_corrects, num_below_one_corrects, \
        num_above_10k_corrects

    


def correct_toa_images(toa_datasets_to_correct_list,
                       atmi_dataset, mask_datasets_list,
                       toa_bands_to_correct_list,
                       atmi_cap_on, atmi_cap_level,
                       nodata_value, cal_slope_model_coeffs_list,
                       cal_offset_model_coeffs_list,
                       max_uint16_val, geotransform_info,
                       proj_info, GTiff_output_image_path,
                       image_basename):

    # Initializations.
    num_bands_to_correct = len(toa_bands_to_correct_list)
    band_num_negative_corrects_list = []
    band_num_below_one_corrects_list = []
    band_num_above_10k_corrects_list = []

    # Get the 10m and 20m image dimensions from the Atm-I
    # images.
    num_img_rows = io.get_num_raster_rows(atmi_dataset)
    num_img_cols = io.get_num_raster_cols(atmi_dataset)

    # Load the 10m Atm-I image data.
    atmi_buffer = io.read_image_from_open_dataset(atmi_dataset)
    if atmi_buffer is None:
        err_msg = f'ERROR: cannot retrieve Atm-I image data from ' \
            + 'open dataset'
        print(err_msg)
        return -1


    ######## MAIN CORRECTION LOOP ##########################################

    for i in range(num_bands_to_correct):

        # Set up to process the current band at its correct spatial resolution
        # and radiometric offset.
        print(f'Correcting band "{toa_bands_to_correct_list[i]}" image...')

        # Get the calibration slope and offset model coefficients for this
        # band.
        band_slope_model_coeffs = cal_slope_model_coeffs_list[i]
        band_offset_model_coeffs = cal_offset_model_coeffs_list[i]

        # Get the TOA image data and mask data for this band.
        toa_image_buffer \
            = io.read_image_from_open_dataset(toa_datasets_to_correct_list[i])
        mask_image_buffer \
            = io.read_image_from_open_dataset(mask_datasets_list[i])

        sr_image_buffer, num_negative_corrects, num_below_one_corrects, \
            num_above_10k_corrects \
            = correct_band_toa(toa_image_buffer, atmi_buffer,
                               mask_image_buffer, num_img_rows,
                               num_img_cols, atmi_cap_on,
                               atmi_cap_level, nodata_value,
                               band_slope_model_coeffs,
                               band_offset_model_coeffs,
                               max_uint16_val)

        band_num_negative_corrects_list.append(num_negative_corrects)
        band_num_below_one_corrects_list.append(num_below_one_corrects)
        band_num_above_10k_corrects_list.append(num_above_10k_corrects)

        # Save the corrected image and get ready for the next band.
        output_image_basename = image_basename + f'_{toa_bands_to_correct_list[i]}_sr'
        status = io.write_image_to_geotiff(sr_image_buffer, proj_info,
                                           geotransform_info, GTiff_output_image_path,
                                           output_image_basename)
        if status < 0:
            err_msg \
                = f'ERROR: cannot save save band {toa_bands_to_correct_list[i]} ' \
                    + 'corrected image data'
            print(err_msg)
            return None, None, None

    ######## END MAIN CORRECTION LOOP ######################################

    # Done.
    return band_num_negative_corrects_list, band_num_below_one_corrects_list, \
        band_num_above_10k_corrects_list
