import os
from pathlib import Path

import cv2
import numpy as np
from osgeo import gdalconst

import cmac_gdal_io as io
import cmac_grid as grid
from cmac_utility import eval_horner_poly



def post_process_raw_atmi(raw_atmi_GTiff_filename,
                          nodata_mask_filename_GTiff_atmi,
                          num_gridcell_rows, num_gridcell_cols,
                          nodata_value, median_filter_on,
                          median_filter_radius, gaussian_filter_on,
                          gaussian_filter_radius, geotransform_info,
                          proj_info, smoothed_atmi_output_path,
                          image_basename):

    # Initialization.
    final_status = 0
    input_image_vector = None
    raw_atmi_grid = None

    final_basename = image_basename + '_atmindex'

    # If we just want to use the "raw" Atm-I image data,
    # copy the raw image to the "smoothed" Atm-I image files.
    if not (median_filter_on and gaussian_filter_on):

        raw_atmi_image_buffer \
            = io.read_image_from_file(raw_atmi_GTiff_filename)
        if not raw_atmi_image_buffer:
            err_msg = 'ERROR: cannot load "raw" Atm-I dataset ' \
                + f'"{raw_atmi_GTiff_filename}" with read-only ' \
                + 'access'
            print(err_msg)
            return -1

        # Get the the image dimensions.
        num_img_rows = raw_atmi_image_buffer.shape[0]
        num_img_cols = raw_atmi_image_buffer.shape[1]

        # Copy the "raw" Atm-I image to the final "smoothed" output
        # file.
        status = io.write_image_to_geotiff(raw_atmi_image_buffer,
                                           proj_info, geotransform_info,
                                           smoothed_atmi_output_path,
                                           final_basename)
        if status < 0:
            err_msg = 'ERROR: cannot create "smoothed" Atm-I image files' \
                + 'from "raw" Atm-I data'
            print(err_msg)
            final_status = -1

    else:

        # Load the "raw" Geo-Tiff ATMI dataset.
        raw_atmi_dataset = io.open_dataset(raw_atmi_GTiff_filename,
                                           io.gdalconst.GA_ReadOnly)
        if not raw_atmi_dataset:
            err_msg = 'ERROR: cannot load "raw" Atm-I dataset ' \
                + f'"{raw_atmi_GTiff_filename}"'
            print(err_msg)
            return -1

        #raw_atmi_raster_band \
        #    = io.get_raster_band_from_dataset(raw_atmi_dataset)
        raw_atmi_image_buffer = io.read_image_from_open_dataset(raw_atmi_dataset)

        # Get the dimensions of the (full-sized) raster image.
        num_img_rows = io.get_num_raster_rows(raw_atmi_dataset)
        num_img_cols = io.get_num_raster_cols(raw_atmi_dataset)
    

        # We've got to apply the specified filter operations on the "raw"
        # ATMI image data. To keep things going relatively quickly, convert
        # the full-sized "raw" Atm-I image to a reduced-size grid image
        # representation, where each grid "pixel" represents a gridcell in
        # the full-sized Atm-I image.
        num_grid_image_rows, num_grid_image_cols, input_image_vector \
            = grid.convert_image_to_vector(raw_atmi_image_buffer,
                                           num_gridcell_rows,
                                           num_gridcell_cols,
                                           nodata_value)
        raw_atmi_grid \
            = grid.convert_vector_to_grid(input_image_vector,
                                          num_grid_image_rows,
                                          num_grid_image_cols,
                                          dtype=input_image_vector.dtype)
        if raw_atmi_grid is None:
            err_msg = 'ERROR: could not create grid image from full-sized ' \
                + 'Atm-I raster image'
            print(err_msg)
            raw_atmi_dataset = None
            return -1

        # Apply the specified smoothing operations to the 'raw' grid image data.
        if median_filter_on:
            median_atmi_grid = cv2.medianBlur(raw_atmi_grid, median_filter_radius)
            if median_atmi_grid is None:
                err_msg = 'ERROR: cannot apply OpenCV median filter to raw grid image'
                print(err_msg)
                raw_atmi_dataset = None
                return -1
        else:
            median_atmi_grid = raw_atmi_grid

        if gaussian_filter_on:
            filter_size = (gaussian_filter_radius, gaussian_filter_radius)
            smoothed_atmi_grid \
                = cv2.GaussianBlur(median_atmi_grid, filter_size, 0)

            if smoothed_atmi_grid is None:
                err_msg = 'ERROR: cannot apply OpenCV gaussian filter to median ' \
                    + 'filtered grid image'
                print(err_msg)
                raw_atmi_dataset = None
                return -1
        else:
            smoothed_atmi_grid = median_atmi_grid

        # Clean up.
        raw_atmi_dataset = None

        # Convert the filtered Atm-I grid image back to a full-sized
        # raster. Start by converting the grid image to a vector
        # representation.
        smoothed_vec = grid.convert_grid_to_vector(smoothed_atmi_grid)
        if smoothed_vec is None:
            err_msg = 'ERROR: cannot convert smoothed grid image to ' \
                + 'vector form'
            print(err_msg)
            return -1

        output_smoothed_raster \
            = grid.convert_vector_to_image(smoothed_vec, num_img_rows,
                                           num_img_cols, num_gridcell_rows,
                                           num_gridcell_cols)

        # Apply the nodata mask to the image.
        nodata_mask_buffer \
            = io.read_image_from_file(nodata_mask_filename_GTiff_atmi)
        if nodata_mask_buffer is None:
            err_msg = 'ERROR: cannot read "nodata" mask image'
            print(err_msg)
            smoothed_atmi_dataset = None

        output_masked_raster = output_smoothed_raster * nodata_mask_buffer

        # Save the smoothed and MPC-scaled Atm-I raster image to file(s).
        status = io.write_image_to_geotiff(output_masked_raster, proj_info,
                                           geotransform_info, smoothed_atmi_output_path,
                                           final_basename)
        #status = io.write_image_to_file(output_masked_raster,
        #                                num_img_rows, num_img_cols,
        #                                geotransform_info, proj_info,
        #                                smoothed_atmi_GTiff_filename,
        #                                smoothed_atmi_product_filename)
        if status < 0:
            err_msg = f'ERROR: cannot create smoothed Atm-I image files'
            print(err_msg)
            final_status = -1
 

    return final_status




def apply_mpc_scaling(smoothed_atmi_filename,
                      mpc_scaling_coeffs,
                      nodata_value,
                      out_image_path,
                      out_image_basename):

    print('Applying MPC scaling to Atm-I prior to CMAC correction...')
    num_mpc_coeffs = len(mpc_scaling_coeffs)

    # Open the smoothed Atm-I image filename.
    atmi_dataset = io.open_dataset(smoothed_atmi_filename, io.gdalconst.GA_ReadOnly)
    if atmi_dataset is None:
        err_msg = f'ERROR: cannot open {smoothed_atmi_filename} to apply MPC scaling'
        print(err_msg)
        return -1

    # Read in the Atm-I data. Get the areas of 'nodata' fill as well.
    atmi_data_buffer = io.read_image_from_open_dataset(atmi_dataset)
    num_img_rows = atmi_data_buffer.shape[0]
    num_img_cols = atmi_data_buffer.shape[1]
    atmi_nodata_pix = (atmi_data_buffer == nodata_value)

    geotransform_info = io.get_geotransform_info(atmi_dataset)
    proj_info_string = io.get_projection_info(atmi_dataset)

    atmi_dataset = None

    # MAIN PROCESSING
    scaled_atmi_data_buffer \
        = np.round((mpc_scaling_coeffs[1]*atmi_data_buffer.astype(np.float64) + mpc_scaling_coeffs[0]), 0)
    scaled_atmi_data_buffer = scaled_atmi_data_buffer.astype(np.uint16)
   

    # Re-establish the 'nodata' fill areas with the proper value.
    scaled_atmi_data_buffer[atmi_nodata_pix] = nodata_value


    # Write the updated Atm-I data buffer out to file.
    image_basename = out_image_basename + '_mpc_scaled'
    status = io.write_image_to_geotiff(scaled_atmi_data_buffer, proj_info_string,
                                       geotransform_info, out_image_path,
                                       image_basename)
    if status < 0:
        tmp_filename = out_image_path + '/' + image_basename + '.tif'
        err_msg = f'ERROR: cannot create MPC-scaled Atm-I dataset {tmp_filename}'
        print(err_msg)

    return status
