import argparse
import datetime as dt
from pathlib import Path
import sys
import time

from osgeo import gdal, gdalconst
import numpy as np

import cmac_config as cfg
import cmac_convert as convert
from cmac_correct import correct_toa_images
import cmac_gdal_io as io
from cmac_generate_raw_atmi import generate_raw_atmi
import cmac_mask as mask
import cmac_metadata as meta
import cmac_postprocess_atmi as post
from cmac_report import generate_summary_report
import cmac_utility as util



def main():

    # Initializations:

    # Required variables.
    cmac_overall_status = 0

    # Set up command-line invocation.
    parser = argparse.ArgumentParser()
    parser.add_argument('cfg_filename')
    cmd_line_args = parser.parse_args()
    cfg_filename = cmd_line_args.cfg_filename
    #cfg_filename = 'C:/Users/tarug/Desktop/cfg_L89_20230904.json'

    # Read the configuration file.
    cfg_param_dict = cfg.read_config(cfg_filename)
    if not cfg_param_dict:
        err_msg = 'ERROR: cannot read configuration file ' \
            + f'"{cfg_filename}"'
        print(err_msg)
        return -1

    # CMAC processing date (YYYYmmdd format)
    cmac_proc_date = dt.datetime.strftime(dt.datetime.now(), '%Y%m%d')
    print(f'PROCESSING DATE:    {cmac_proc_date}')

    # Start timing CMAC processing.
    cmac_start_time = time.perf_counter()


    ################ CMAC INITIALIZATION ####################################

    # Maximum 16-bit unsigned integer value used to check whether Atm-I and/or
    # CMAC corrected pixel values are "out-of-range"
    max_uint16_val = np.iinfo(np.uint16).max

    # Start pulling configuration info:

    # Imaging system info.
    satellite = cfg_param_dict["satellite"]
    sensor = cfg_param_dict["sensor"]

    # Path and file-related info.
    root_toa_path = cfg_param_dict["root_toa_path"]
    root_cmac_path = cfg_param_dict["root_cmac_path"]
    additional_path_info = cfg_param_dict["additional_path_info"]
    product_id = cfg_param_dict["product_id"]
    file_ext = cfg_param_dict["file_ext"]
    option1_folder = cfg_param_dict["option1_folder"]
    option2_folder = cfg_param_dict["option2_folder"]

    # Atm-I related info.
    atmi_model_coeffs = cfg_param_dict["atmi_model_coeffs"]
    atmi_gridcell_size = cfg_param_dict["atmi_gridcell_size"]

    median_filter_on = cfg_param_dict["median_filter_on"]
    median_filter_radius = cfg_param_dict["median_filter_radius"]
    gaussian_filter_on = cfg_param_dict["gaussian_filter_on"]
    gaussian_filter_radius = cfg_param_dict["gaussian_filter_radius"]

    atmi_cap_on = cfg_param_dict["atmi_cap_on"]
    atmi_cap_level = cfg_param_dict["atmi_cap_level"]

    # Sensor-specific MPC scaling. Apply to Atm-I prior to correction.
    # Specified in reverse polynomial order (i.e. constant, linear term,
    # etc.).
    mpc_scaling_coeffs = [cfg_param_dict["mpc_scaling_offset"], \
        cfg_param_dict["mpc_scaling_gain"]]

    # Calibration model coefficient sets.
    cal_offset_models_list = cfg_param_dict["cal_offset_models"]
    cal_slope_models_list = cfg_param_dict["cal_slope_models"]

    # General processing info.
    nodata_value = cfg_param_dict["nodata_value"]
    bands_for_atmi_model = cfg_param_dict["bands_for_atmi_model"]
    bands_to_correct = cfg_param_dict["bands_to_correct"]

    # CMAC version-related info.
    cmac_cal_version = cfg_param_dict["cmac_cal_version"]
    cmac_software_version = cfg_param_dict["cmac_software_version"]
    if not (cmac_cal_version or cmac_software_version):
        err_msg = 'ERROR: CMAC calibration model and/or software version ' \
            + 'info not specified'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print('TOTAL PROCESSING TIME:  {elapsed_time:.2f} seconds')
        return -1

    # Make sure the 'root' TOA image path exists.
    if not Path(root_toa_path).exists():
        err_msg = f'ERROR: Root TOA image path "{root_toa_path}" not found'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME:  {elapsed_time:.2f} seconds')
        return -1

    # Get the final path to the TOA image data (before conversion to
    # scaled L8 TOAR reflectance). Once constructed, check to
    # ensure it exists.
    final_toa_path = root_toa_path + '/' + satellite + '/' + sensor + '/' \
        + additional_path_info + '/' + product_id
    if not Path(final_toa_path).exists():
        err_msg = f'ERROR: Full TOA image path "{final_toa_path}" not found'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME:  {elapsed_time:.2f} seconds')
        return -1

    # Construct the final path to the CMAC image data. Once constructed,
    # check to make sure it exists, otherwise create it.
    final_cmac_path = root_cmac_path + '/' + satellite + '/' + sensor + '/' \
        + additional_path_info + '/' + 'CMAC' + '/' + cmac_cal_version \
        + '/' + cmac_software_version + '/' + cmac_proc_date + '/' \
        + product_id
 
    if option1_folder != 'NONE':
        final_cmac_path = final_cmac_path + '/' + option1_folder
    if option2_folder != 'NONE':
        final_cmac_path = final_cmac_path + '/' + option2_folder

    if not Path(final_cmac_path).exists():
        print(f'CMAC output path "{final_cmac_path}" not found, creating...')
        out_path = Path(final_cmac_path)
        out_path.mkdir(parents=True, exist_ok=True)

    # If it does exist, clean out the file contents.
    else:
        print(f'CMAC output path "{final_cmac_path}" found, deleting previous contents...')
        util.clean_cmac_gtiffs(final_cmac_path)


    # Get the ID strings for the band(s) to be used in Atm-I estimation
    # and final correction. Then use these to construct the corresponding
    # set of TOA input and CMAC output filenames.
    num_bands_to_model = len(bands_for_atmi_model)
    num_bands_to_correct = len(bands_to_correct)

    toa_bands_to_model_list = []
    for b in bands_for_atmi_model:
        toa_bands_to_model_list.append(f'B{b}')

    toa_bands_to_correct_list = []
    for b in bands_to_correct:
        toa_bands_to_correct_list.append(f'B{b}')

    if not (toa_bands_to_model_list and toa_bands_to_correct_list):
        err_msg = 'ERROR: could not retrieve band ID list(s) for Atm-I ' \
            + 'modeling and/or correction'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME:  {elapsed_time:.2f} seconds')
        return -1

    toa_atmi_image_filenames \
        = util.construct_image_filenames(final_toa_path,
                                         product_id,
                                         '', '',
                                         toa_bands_to_model_list,
                                         file_ext)
    toa_correct_image_filenames \
        = util.construct_image_filenames(final_toa_path,
                                         product_id,
                                         '', '',
                                         toa_bands_to_correct_list,
                                         file_ext)
    atmi_nodata_mask_image_filenames \
        = util.construct_image_filenames(final_cmac_path,
                                         product_id,
                                         '', 'nodata_mask',
                                         toa_bands_to_model_list,
                                         file_ext)
    corrected_nodata_mask_image_filenames \
        = util.construct_image_filenames(final_cmac_path,
                                         product_id,
                                         '', 'nodata_mask',
                                         toa_bands_to_correct_list,
                                         file_ext)
    cmac_corrected_image_filenames \
        = util.construct_image_filenames(final_cmac_path,
                                         product_id,
                                         '', 'sr',
                                         toa_bands_to_correct_list,
                                         file_ext)

    # Construct the Atm-I image filenames ('raw' and 'smoothed').
    smoothed_atmi_GTiff_filename = final_cmac_path + '/' + product_id \
        + f'_atmindex.{file_ext}'
    raw_atmi_GTiff_filename = final_cmac_path + '/' + product_id \
        + f'_raw_atmindex.{file_ext}'

    # Select one of the correction files as the 'reference' file
    # for projection and geometric transformation info. Any file
    # would do since all have the same spatial resolution. We'll
    # use the first file in the list. Then get the geometric
    # projection and transformation info.
    georeference_filename = toa_atmi_image_filenames[0]
    georef_ds = io.open_dataset(georeference_filename, gdalconst.GA_ReadOnly)
    if georef_ds is None:
        err_msg = 'ERROR: cannot open geo-reference image file ' \
            + f'"{georeference_filename}"'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME:  {elapsed_time:.2f} seconds')
        return -1

    geotransform_info = io.get_geotransform_info(georef_ds)
    proj_info = io.get_projection_info(georef_ds)
    georef_ds = None

    if geotransform_info is None or proj_info is None:
        err_msg = 'ERROR: cannot retrieve geometric transform and/or ' \
            + f'projection information from "{georeference_filename}"'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print('TOTAL PROCESSING TIME:  {elapsed_time:.2f} seconds')
        return -1

    ################ END CMAC INITIALIZATION ################################


    ################ METADATA INITIALIZATION ################################

    # Get the path to the TOA metadata file, then retrieve the relevant
    # processing metadata (i.e. solar zenith angle and band-wise reflectance
    # conversion coeffficients). Once we have the solar zenith angle,
    # determine the scaling factor to apply to the converted pixel values to
    # get them into Sentinel-2 "space". Make sure to periodically keep
    # tabs on the Sentinel-2 quantification value (which is currently
    # represented here by the '10000.' value; it can be found in the
    # product's "MSIL1C_MTD.xml" metadata file located in the main
    # ".SAFE" folder.
    print('Retrieving processing metadata...')
    metadata_filename = final_toa_path + '/' + product_id + '_MTL.xml'

    solar_zenith_angle = meta.get_solar_zenith_angle(metadata_filename)
    cos_theta = np.cos((np.pi / 180) * solar_zenith_angle)
    magnitude_scaling_factor = 10000. / cos_theta


    atmi_band_reflectance_conversion_factors \
        = meta.get_band_reflectance_conversion_factors(metadata_filename,
                                                       toa_bands_to_model_list)
    correction_band_reflectance_conversion_factors \
        = meta. get_band_reflectance_conversion_factors(metadata_filename,
                                                        toa_bands_to_correct_list)

    # Debugging purposes only!
    #print('Atm-I BAND REFLECTANCE CONVERSION FACTORS:')
    #print(atmi_band_reflectance_conversion_factors)
    #print('')
    #print('CORRECTION BAND REFLECTANCE CONVERSION FACTORS:')
    #print(correction_band_reflectance_conversion_factors)
    #print('')
    #print(f'SOLAR ZENITH ANGLE:     {solar_zenith_angle}')

    ################ END METADATA INITIALIZATION ############################


    ################ GDAL INITIALIZATION ####################################

    # Make sure the required GDAL GeoTiff image driver is installed.
    # We won't be using this driver object directly in any processing --
    # we're just testing for its existence, and aborting further processing
    # if for some reason it can't be found.
    #
    # NOTE: GDALAllRegister(), which initializes ALL of GDAL's installed
    #       image drivers, does not have to be called from the main function
    #       in Python -- it is automatically called when the 'gdal' module
    #       is imported. Nice change from C/C++, where it does have to be
    #       explicitly called before calling on any other GDAL functionality.
    print('Locating required image drivers...')
    test_GTiff_driver = gdal.GetDriverByName('GTiff')
    if not test_GTiff_driver:
        err_msg = 'ERROR: "GeoTiff" image driver not found, may not be ' \
            + 'installed. Check GDAL documentation for the correct driver ' \
            + 'name.'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME:  {elapsed_time:.2f} seconds')
        return -1
    else:
        print('"GeoTiff" driver found in processing system GDAL installation')

    ################ END GDAL INITIALZATION #################################


    ################ LOAD / CONVERT Atm-I MODELING BANDS ####################

    # Convert the image(s) for Atm-I estimation.
    converted_atmi_datasets = []
    num_bands_converted = 0
    for index, fname in enumerate(toa_atmi_image_filenames):

        conversion_factors = atmi_band_reflectance_conversion_factors[index]
        converted_fname = convert.convert_toa_band(fname, conversion_factors,
                                                   nodata_value,
                                                   magnitude_scaling_factor,
                                                   proj_info,
                                                   geotransform_info,
                                                   final_cmac_path,
                                                   product_id)

        # Prepare the converted datasets for processing.
        if converted_fname != '':
            ds = io.open_dataset(converted_fname, gdalconst.GA_ReadOnly)
            if ds is not None:
                converted_atmi_datasets.append(ds)
                num_bands_converted += 1
     
    if num_bands_converted != num_bands_to_model:
        err_msg = f'ERROR: {num_bands_converted} Atm-I bands converted, ' \
            + f'but {num_bands_to_model} needed to be converted'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME:  {elapsed_time:.2f} seconds')
        if converted_atmi_datasets:
            for ds in converted_atmi_datasets:
                ds = None
        return -1

    ################ END LOAD / CONVERT Atm-I MODELING BANDS ################


    ################ GENERATE Atm-I BAND 'NODATA' MASKS #####################

    print('Generating "nodata" mask images for Atm-I input bands...')
    status = mask.create_nodata_mask_images(converted_atmi_datasets,
                                            toa_bands_to_model_list,
                                            nodata_value,
                                            proj_info,
                                            geotransform_info,
                                            final_cmac_path,
                                            product_id)
    if status < 0:
        err_msg = 'ERROR: failed to create all required "nodata" mask images ' \
            + f'for Atm-I estimation bands'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print('TOTAL PROCESSING TIME:  {elapsed_time:.2f} seconds')
        for ds in converted_atmi_datasets:
            ds = None
        return -1

    ################ END GENERATE Atm-I BAND 'NODATA' MASKS #################



    ################ GENERATE RAW Atm-I #####################################

    print('Generating (raw) Atm-I image from VNIR band inputs...')

    print(f'Atm-I GRIDCELL SIZE (pixels):  {atmi_gridcell_size[0]}')

    # Time how long it takes to create the final Atm-I image. This
    # will include generating the raw image and performing any/all
    # specified smoothing operations on it.
    atmi_start_time = time.perf_counter()

    # Create the raw Atm-I dataset.
    atmi_model_coeffs = cfg_param_dict["atmi_model_coeffs"]
    status \
        = generate_raw_atmi(converted_atmi_datasets, toa_bands_to_model_list,
                            nodata_value, atmi_gridcell_size, atmi_model_coeffs,
                            max_uint16_val, proj_info, geotransform_info,
                            atmi_nodata_mask_image_filenames[0],
                            final_cmac_path, product_id)

    # Free up resources by "closing" the open dataset(s).
    for ds in converted_atmi_datasets:
        ds = None

    if status < 0:
        err_msg = 'ERROR: could not create "raw" 10-m Atm-I dataset'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        util.clean_cmac_directory(final_cmac_path)
        return -1

    ################ END GENERATE RAW Atm-I #################################


    ################ APPLY Atm-I SMOOTHING ##################################

    # Atm-I smoothing options.
    median_filter_on = cfg_param_dict["median_filter_on"]
    median_filter_radius = cfg_param_dict["median_filter_radius"]
    gaussian_filter_on = cfg_param_dict["gaussian_filter_on"]
    gaussian_filter_radius = cfg_param_dict["gaussian_filter_radius"]
    

    # For L8/L9, apply any MPC scaling factor here, so that the final
    # Atm-I output is scaled.
    status = post.post_process_raw_atmi(raw_atmi_GTiff_filename,
                                        atmi_nodata_mask_image_filenames[0],
                                        atmi_gridcell_size[0], atmi_gridcell_size[1],
                                        nodata_value, median_filter_on,
                                        median_filter_radius, gaussian_filter_on,
                                        gaussian_filter_radius, geotransform_info,
                                        proj_info, final_cmac_path,
                                        product_id)
    if status < 0:
        err_msg = 'ERROR: generation of "smoothed" atm-i datasets failed'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME:  {elapsed_time:.2f} seconds')
        util.clean_cmac_directory(final_cmac_path)
        return -1

    ################ END APPLY Atm-I SMOOTHING ##############################



    ################ APPLY MPC SCALING TO Atm-I #############################

    smoothed_atmi_filename = final_cmac_path + '/' + product_id + '_atmindex.tif'
    mpc_atmi_basename = product_id + '_atmindex'
    status = post.apply_mpc_scaling(smoothed_atmi_filename, mpc_scaling_coeffs,
                                    nodata_value, final_cmac_path,
                                    mpc_atmi_basename)
    if status < 0:
        err_msg = 'ERROR: cannot scale Atm-I for MPC processing'
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'total processing time: {elapsed_time:.2f} seconds')
        util.clean_cmac_directory(final_cmac_path)
        return -1

    ################ END APPLY MPC SCALING TO Atm-I #########################

    atmi_stop_time = time.perf_counter()
    elapsed_atmi_time = atmi_stop_time - atmi_start_time
    print(f'TOTAL Atm-I PROCESSING TIME:  {elapsed_atmi_time:.2f} seconds')



    # PART 2 -- APPLY CMAC CORRECTION

    print('Initializing CMAC correction processing:')

    # Initializations:

    # Calibration slope and offset models.
    band_cal_slope_models_list = cfg_param_dict["cal_slope_models"]
    band_cal_offset_models_list = cfg_param_dict["cal_offset_models"]

    # Atm-I capping and cap level.
    atmi_cap_on = cfg_param_dict["atmi_cap_on"]
    atmi_cap_level = cfg_param_dict["atmi_cap_level"]

    # Load up the Atm-I dataset.
    print('Loading Atm-I image for correction...')
    mpc_scaled_atmi_filename = final_cmac_path + '/' + product_id \
        + '_atmindex_mpc_scaled.tif'
    smoothed_atmi_dataset = io.open_dataset(mpc_scaled_atmi_filename,
                                            gdalconst.GA_ReadOnly)
    if not smoothed_atmi_dataset:
        err_msg = f'ERROR: cannot open {smoothed_atmi_GTiff_filename} with ' \
            + 'read-only access'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME:  {elapsed_time:.2f} seconds')
        for ds in converted_correct_datasets:
            ds = None
        util.clean_cmac_directory(final_cmac_path)
        return -1

    ################### LOAD/CONVERT CMAC BANDS TO CORRECT ##################

    converted_correct_datasets = []
    num_bands_converted = 0
    for index, fname in enumerate(toa_correct_image_filenames):

        conversion_factors = correction_band_reflectance_conversion_factors[index]
        converted_fname = convert.convert_toa_band(fname, conversion_factors,
                                                   nodata_value,
                                                   magnitude_scaling_factor,
                                                   proj_info,
                                                   geotransform_info,
                                                   final_cmac_path,
                                                   product_id)

        # Prepare the converted datasets for processing.
        if converted_fname != '':
            ds = io.open_dataset(converted_fname, gdalconst.GA_ReadOnly)
            if ds is not None:
                converted_correct_datasets.append(ds)
                num_bands_converted += 1
     
    if num_bands_converted != num_bands_to_correct:
        err_msg = f'ERROR: {num_bands_converted} bands-to-correct converted, ' \
            + f'but {num_bands_to_model} needed to be converted'
        print(err_msg)
        if converted_correct_datasets:
            for ds in converted_correct_datasets:
                ds = None
        smoothed_atmi_dataset = None
        util.clean_cmac_directory(final_cmac_path)
        cmac_stop_time = time.perf_counter()
        for ds in converted_correct_datasets:
            ds = None
        return -1
 
 
    ################# END LOAD/CONVERT CMAC BANDS TO CORRECT ################


    ################# GENERATE/LOAD CORRECTION 'NODATA' MASKS ###############

    print('Generating "nodata" mask images for input bands to correct...')
    status = mask.create_nodata_mask_images(converted_correct_datasets,
                                            toa_bands_to_correct_list,
                                            nodata_value,
                                            proj_info,
                                            geotransform_info,
                                            final_cmac_path,
                                            product_id)
    if status < 0:
        err_msg = 'ERROR: failed to create all required "nodata" mask images ' \
            + f'for Atm-I estimation bands'
        print(err_msg)
        for ds in converted_correct_datasets:
            ds = None
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print('TOTAL PROCESSING TIME:  {elapsed_time:.2f} seconds')
        util.clean_cmac_directory(final_cmac_path)
        return -1

    # Load the correction band 'nodata' mask images.
    correction_mask_datasets = []
    num_masks_loaded = 0
    num_masks_to_load = len(corrected_nodata_mask_image_filenames)

    for index, fname in enumerate(corrected_nodata_mask_image_filenames):
        ds = io.open_dataset(fname, gdalconst.GA_ReadOnly)
        if not ds:
            err_msg = f'ERROR: cannot load {fname}'
            print(err_msg)
            for image_ds in converted_correct_datasets:
                image_ds = None
            if corrected_mask_datasets:
                for image_ds in corrected_mask_datasets:
                    image_ds = None
            smoothed_atmi_dataset = None
            cmac_stop_time = time.perf_counter()
            elapsed_time = cmac_stop_time - cmac_start_time
            print('TOTAL PROCESSING TIME:  {elapsed_time:.2f} seconds')
            util.clean_cmac_directory(final_cmac_path)
            return -1

        correction_mask_datasets.append(ds)
        num_masks_loaded += 1

    if num_masks_loaded != num_masks_to_load:
        err_msg = f'ERROR: {num_masks_loaded} mask images loaded, but ' \
            + f'{num_masks_to_load} needed to be loaded'
        print(err_msg)
        for ds in correction_mask_datasets:
            ds = None
        for ds in converted_correct_datasets:
            ds = None
        smoothed_atmi_dataset = None
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print('TOTAL PROCESSING TIME:  {elapsed_time:.2f} seconds')
        util.clean_cmac_directory(final_cmac_path)
        return -1

    ############# END GENERATE/LOAD CORRECTION 'NODATA' MASKS ###############


    ################# CORRECT TOA IMAGE BANDS ###############################

    # Time the correction step.
    correct_start_time = time.perf_counter()
    correct_stop_time = correct_start_time

    print('Correcting TOA images...')
    cmac_negative_correct_errors_list, cmac_below_one_correct_errors_list, \
        cmac_above_10k_correct_errors_list \
        = correct_toa_images(converted_correct_datasets, smoothed_atmi_dataset,
                             correction_mask_datasets, toa_bands_to_correct_list,
                             atmi_cap_on, atmi_cap_level, nodata_value,
                             cal_slope_models_list, cal_offset_models_list,
                             max_uint16_val, geotransform_info,
                             proj_info, final_cmac_path, product_id)

    # If we've made it out of the correction stage, stop timing it.
    correct_stop_time = time.perf_counter()
    elapsed_correct_time = correct_stop_time - correct_stop_time

    # Close up shop before final error/status checking.
    for ds in converted_correct_datasets:
        ds = None
    for ds in correction_mask_datasets:
        ds = None
    smoothed_atmi_dataset = None

    if not (cmac_negative_correct_errors_list and cmac_below_one_correct_errors_list \
        and cmac_above_10k_correct_errors_list):
        err_msg = 'ERROR: Correction processing failed -- no correction error ' \
            + 'statistics generated'
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print('TOTAL PROCESSING TIME:  {elapsed_time:.2f} seconds')
        util.clean_cmac_directory(final_cmac_path)
        return -1

    ################# END CORRECT TOA IMAGE BANDS ###########################

    
    ################# GENERATE SUMMARY REPORT ###############################

    # Stop processing timing -- generating the final report shouldn't really
    # count here.
    cmac_stop_time = time.perf_counter()
    elapsed_proc_time = cmac_stop_time - cmac_start_time
    print(f'TOTAL PROCESSING TIME: {elapsed_proc_time:.2f} seconds')

    # Generate the final report.
    print('Generating processing summary report...')
    cmac_cal_version = cfg_param_dict["cmac_cal_version"]
    cmac_software_version = cfg_param_dict["cmac_software_version"]
    status = generate_summary_report(smoothed_atmi_GTiff_filename,
                                     nodata_value, toa_bands_to_correct_list,
                                     elapsed_atmi_time, elapsed_correct_time,
                                     elapsed_proc_time,
                                     cmac_negative_correct_errors_list,
                                     cmac_below_one_correct_errors_list,
                                     cmac_above_10k_correct_errors_list,
                                     final_cmac_path, product_id,
                                     cmac_proc_date, cmac_cal_version,
                                     cmac_software_version)
    if status < 0:
        err_msg = 'ERROR: could not generate summary report'
        print(err_msg)
        cmac_overall_status = -1

    ################# END GENERATE SUMMARY REPORT ###########################

    return cmac_overall_status






if __name__ == '__main__':
    main()
